/* ksnet.cc
----------------------------------------------------------------
   饹 Ksnet μ¸
----------------------------------------------------------------
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "common.h"                                 /* ̴Ķ */
#include "list.h"                               /* 饹 List  */
#include "element.h"                         /* 饹 Element  */
#include "p_tmp.h"         /* ƥץ졼ȥ饹 Receive, Send  */
#include "ksnet.h"                             /* 饹 Ksnet  */


Ksnet::Ksnet(char* o_name, int o_num, char* p_str,
	     int numinput, int numlateral, int lateral_unit_tmp,
	     int numsegment, int numoutput, int t_step)
: Element()
{
  class_name = "Ksnet";
  Init(o_name, o_num, p_str, 
       numinput, numlateral, lateral_unit_tmp, 
       numsegment, numoutput, t_step);
}


Boolean Ksnet::Init(char* o_name, int o_num, char* p_str,
		    int numinput, int numlateral, int lateral_unit_tmp, 
		    int numsegment, int numoutput, int t_step)
{
  /* tStep */
  if (t_step <= 0){
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'tStep' must be natural number.
    Oprint();
    exit(1);
  }
  tStep = t_step;

  /* ninput */
  if (numinput <= 0) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'ninput' must be natural number.
    Oprint();
    exit(1);
  }
  ninput = numinput;
  
  /* nlateral */
  if (numlateral < 0) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'nlateral' must not be negative number.
    Oprint();
    exit(1);
  }
  nlateral = numlateral;

  /* lateral_unit */
  if ((lateral_unit_tmp != 2) && (lateral_unit_tmp != 3)){
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'lateral_unit' must be "2(m^2/sec)" or "3(m^3/sec)".
    Oprint();
    exit(1);
  }
  lateral_unit = lateral_unit_tmp;
  
  /* nsegment */
  if (numsegment <= 0) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'nsegment' must be natural number.
    Oprint();
    exit(1);
  }
  nsegment = numsegment;
  
  /* noutput */
  if (numoutput <= 0) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // 'noutput' must be natural number.
    Oprint();
    exit(1);
  }
  noutput = numoutput;

  /* üҡüҤν */
  if ((rpl = new Receive<double>[nlateral]) == NULL) {
    Eprint_memory(__FILE__, __LINE__);
    // Ksnet::Init
    // Cannot allocate memory for rpl.
    Oprint();
    exit(1);
  }
  int i;
  for (i = 0; i < nlateral; i++) {
    rpl[i].Init("rpl", i);
  }
  
  if ((rpu = new Receive<double>[ninput]) == NULL) {
    Eprint_memory(__FILE__, __LINE__);
    // Ksnet::Init
    // Cannot allocate memory for rpu.
    Oprint();
    exit(1);
  }
  for (i = 0; i < ninput; i++) {
    rpu[i].Init("rpu", i);
  }
  
  if ((sp = new Send<double>[noutput]) == NULL) {
    Eprint_memory(__FILE__, __LINE__);
    // Ksnet::Init
    // Cannot allocate memory for sp.
    Oprint();
    exit(1);
  }
  for (i = 0; i < noutput; i++) {
    sp[i].Init("sp", i);
  }

  /* ե̾ */
  char fn_sub[50];
  Set_data_string(p_str);
  if (sscanf(data_string, "%s", fn_sub) != 1) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Init
    // Cannot scan filename for Sub_connect.
    Oprint();
    exit(1);
  }
  Clear_data_string();

  /* ƻ³ */
  Sub_connect(fn_sub);

  return Element::Init(o_name, o_num);
}


Ksnet::Ksnet(void) : Element()
{
  class_name = "Ksnet";
}


Ksnet::~Ksnet(void)
{
  delete [] L;
  delete [] B;
  delete [] G;
  delete [] K;
  delete [] P;
  delete [] alpha;

  delete [] seg_NO;
  for (int i = 0; i < nsegment; i++) {
    delete [] mu_up[i];
  }
  delete [] mu_up;
  delete [] sigma_up;
  delete [] sigma_down;
  for (int i = 0; i < nsegment; i++) {
    delete [] mu_down[i];
  }
  delete [] mu_down;
  delete [] rpu_NO;
  delete [] sp_NO;
  delete [] cal_order;

  delete [] Q_up;
  delete [] Q_down;
  
  delete [] rpu;
  delete [] rpl;
  delete [] sp;
}


void Ksnet::Register_receive_ports(void)
{
  int i;
  for (i = 0; i < nlateral; i++) {
    Register((void*) &rpl[i]);
  }
  for (i = 0; i < ninput; i++) {
    Register((void*) &rpu[i]);
  }
}


void Ksnet::Register_send_ports(void)
{
  int i;
  for (i = 0; i < noutput; i++) {
    Register((void*) &sp[i]);
  }
}


void Ksnet::Set_parameter(FILE* fp)
{
  L = new double[nsegment];
  B = new double[nsegment];
  G = new double[nsegment];
  K = new double[nsegment];
  P = new double[nsegment];
  alpha = new double[nsegment];

  int i, j;
  for (i = 0; i < nsegment; i++) {
    L[i] = -99999.; B[i] = -99999.; G[i] = -99999.;
    K[i] = -99999.; P[i] = -99999.; alpha[i] = -99999.;
  }

  int n_tmp;
  double l_tmp, b_tmp, g_tmp, m_tmp, p_tmp;
  for (i = 0; i < nsegment; i++) {
    if (fscanf(fp, "%d %lf %lf %lf %lf %lf",
	       &n_tmp, &l_tmp, &b_tmp, &g_tmp, &m_tmp, &p_tmp) != 6) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_parameter
      // Cannot scan information of parameter.
      Oprint();
      exit(1);
    }

    if ((n_tmp <= 0) || (l_tmp < 0.) || (b_tmp <= 0.) || 
	(g_tmp <= 0.) || (m_tmp <= 0.) || (p_tmp <= 0.)) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_parameter
      // Invalid parameter are given.
      Oprint();
      exit(1);
    }
    if (p_tmp > 1.) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_parameter
      // 'P' must be smaller than 1.
      Oprint();
      exit(1);
    }

    for(j = 0; j < nsegment; j++){
      if(seg_NO[j] == n_tmp){
	L[j] = l_tmp;
	B[j] = b_tmp;
	G[j] = g_tmp;
	K[j] = pow(b_tmp, 1. - p_tmp)
	  * pow((sqrt(g_tmp) / m_tmp), - p_tmp);
	P[j] = p_tmp;

//	printf("seg_NO = %d, L[%d] = %f\n", seg_NO[j], j, L[j]);

	break;
      }
    }
  }

  /* ήʬΨ alpha η׻ */
  int tmp;
  double BsqrtG, sum_BsqrtG;
  for(i = 0; i < nsegment; i++){
    if (sigma_up[i] == 0) {
      alpha[i] = 1.;
//      printf("seg_NO = %d, alpha[%d] = %f \n", seg_NO[j], i, alpha[i]);
    }
    else{
      sum_BsqrtG = B[i] * sqrt(G[i]);
      for(j = 0; j < sigma_up[i]; j++){
	if(mu_up[i][j] < 0){
	  tmp = abs(mu_up[i][j]) - 1;
	  BsqrtG = B[tmp] * sqrt(G[tmp]);
	  sum_BsqrtG += BsqrtG;
	}
      }
      alpha[i] = (B[i] * sqrt(G[i])) / sum_BsqrtG;
//      printf("seg_NO = %d, alpha[%d] = %f \n", seg_NO[i], i, alpha[i]);
    }
  }

  /* sum_length  area η׻ */
  sum_length = 0.; area = 0.;
  for (i = 0; i < nsegment; i++) {
    sum_length += L[i];
    area += L[i] * B[i];
  }

  /* γ */
  Q_up = new double[nsegment];
  Q_down = new double[nsegment];
}


void Ksnet::Set_initial_state(FILE* fp)
{
  /* ׻ϻ */
  char s[256];
  if (fgets(s, 256, fp) == NULL) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Set_initial_state
    // Cannot read 'current_time'.
    Oprint();
    exit(1);
  }
  if (sscanf(s, "%ld", &current_time) != 1) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Set_initial_state
    // Cannot get 'current_time'.
    Oprint();
    exit(1);
  }

  int i;
  char** buf = new char*[3];
  for (i = 0; i < 3; i++) {
    buf[i] = new char[256];
    if (fgets(buf[i], 256, fp) == NULL) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // Cannot get data.
      Oprint();
      exit(1);
    }
  }

  /* ή */
  double Q_tmp;
  sum_input_current = 0.;
  for (i = 0; i < ninput; i++) {
    // ڡ򥹥å
    while ((*buf[0] == ' ') || (*buf[0] == '\t')) buf[0]++;
    if (sscanf(buf[0], "%lf", &Q_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // Cannot get 'inflow'.
      Oprint();
      exit(1);
    }
    // 餽ʬʤƤ
    while ((*buf[0] != ' ') && (*buf[0] != '\t')) buf[0]++;
    
    if (Q_tmp < 0.) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // 'inflow' should NOT be a negative value.
      Oprint();
      exit(1);
    }
    Q_up[rpu_NO[i]] = Q_tmp;
    sum_input_current += Q_tmp;
  }
  
  /* ¦ή */
  lateral_current_time = 0.;
  for (i = 0; i < nlateral; i++) {
    // ڡ򥹥å
    while ((*buf[1] == ' ') || (*buf[1] == '\t')) buf[1]++;
    if (sscanf(buf[1], "%lf", &Q_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // Cannot get 'lateral inflow'.
      Oprint();
      exit(1);
    }
    // 餽ʬʤƤ
    while ((*buf[1] != ' ') && (*buf[1] != '\t')) buf[1]++;
    
    if (Q_tmp < 0.) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // 'lateral inflow' should NOT be a negative value.
      Oprint();
      exit(1);
    }
    lateral_current_time += Q_tmp;
  }

  /* ή */
  sum_output_current = 0.;
  for (i = 0; i < noutput; i++) {
    // ڡ򥹥å
    while ((*buf[2] == ' ') || (*buf[2] == '\t')) buf[2]++;
    if (sscanf(buf[2], "%lf", &Q_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // Cannot get 'outflow'.
      Oprint();
      exit(1);
    }
    // 餽ʬʤƤ
    while ((*buf[2] != ' ') && (*buf[2] != '\t')) buf[2]++;
    if (Q_tmp < 0.) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Set_initial_state
      // 'outflow' should NOT be a negative value.
      Oprint();
      exit(1);
    }
    Q_down[sp_NO[i]] = Q_tmp;
    sum_output_current += Q_tmp;
  }

  for (i = 0; i < 3; i++) {
    buf[i] = NULL;
    delete [] buf[i];
  }
  delete [] buf;

}


void Ksnet::Save_terminal_state(FILE* fp)
{
  if (fp == NULL) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Save_terminal_state
    // Cannot save terminal state.
    Oprint();
    exit(1);
  }

  /* current_time */
  fprintf(fp, "%ld\n", current_time);
  
  /* ǾήüǤή */
  int i;
  for (i = 0; i < ninput; i++)
    fprintf(fp, "%lf ", Q_up[rpu_NO[i]]);
  fprintf(fp, "\n");
  
  /* ¦ή */
  double tmp;
  for (i = 0; i < nlateral; i++) {
    tmp = rpl[i].Get_data(current_time);
    fprintf(fp, "%lf ", tmp);
  }
  fprintf(fp, "\n");

  /* ǲήüǤή */
  for (i = 0; i < noutput; i++)
    fprintf(fp, "%lf ", Q_down[sp_NO[i]]);
  fprintf(fp, "\n"); 
}


void Ksnet::Sub_connect(char* cf_name)
{
  FILE* fp_c;
  if ((fp_c = fopen(cf_name, "r")) == NULL) {
    Eprint_fopen(__FILE__, __LINE__, cf_name, "r");
    Oprint();
    exit(1);
  }

  seg_NO = new int[nsegment];
  sigma_up = new int[nsegment];
  mu_up = new int*[nsegment];
  sigma_down = new int[nsegment];
  mu_down = new int*[nsegment];
  cal_order = new int[nsegment];
  rpu_NO = new int[ninput];
  sp_NO = new int[noutput];

  int** MU_up;
  MU_up = new int*[nsegment];
  int** MU_down;
  MU_down = new int*[nsegment];

  char* buf = new char[256];
  int line = 0; int segment = -1; int i, j;         // 󥿡

  int seg_NO_tmp, sigma_up_tmp, MU_up_tmp; 
  int sigma_down_tmp, MU_down_tmp;

  char* pbuf;

  while (fgets(buf, 256, fp_c) != NULL) {
    line++;
    pbuf = buf;
    if ((buf[0] == '#') || (buf[0] == '\n')) continue;
    segment++;
    if (segment == nsegment) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // Too many information for sub-connect.
      Oprint();
      exit(1);
    }
    
    if (sscanf(pbuf, "%d", &seg_NO_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // Cannot scan 'seg_NO'.
      fprintf(stderr, "\tLine %d.\n", line);
      Oprint();
      exit(1);
    }
    if (seg_NO_tmp <= 0) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // 'seg_NO' must be a natural number.
      fprintf(stderr, "\tLine %d.\n", line);
      Oprint();
      exit(1);
    }
    seg_NO[segment] = seg_NO_tmp;

    while (*pbuf != ':') pbuf++;
    pbuf++;

    if (sscanf(pbuf, "%d", &sigma_up_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // Cannot scan 'sigma_up'.
      fprintf(stderr, "\tLine %d.\n", line);
      Oprint();
      exit(1);
    }
    if (sigma_up_tmp < 0) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // 'sigma_up' should NOT have negative value.
      fprintf(stderr, "\tLine %d.\n", line);
      Oprint();
      exit(1);
    }
    sigma_up[segment] = sigma_up_tmp;

    while (*pbuf != ':') pbuf++;
    pbuf++;

    if (sigma_up[segment] > 0){
      MU_up[segment] = new int[sigma_up[segment]];
      for (j = 0; j < sigma_up[segment]; j++) {
	while ((*pbuf == ' ') || (*pbuf == '\t')) pbuf++;
	if (sscanf(pbuf, "%d", &MU_up_tmp) != 1) {
	  Eprint(__FILE__, __LINE__);
	  // Ksnet::Sub_connect
	  // Cannot scan 'MU_up'.
	  Oprint();
	  exit(1);
	}
	MU_up[segment][j] = MU_up_tmp;  // MU_up ˤϼºݤβƻʬֹ椬
		          		// Ǽ롣   

	while ((*pbuf != ' ') && (*pbuf != '\t')) pbuf++;
      }
    }

    else {
      while ((*pbuf == ' ') || (*pbuf == '\t')) pbuf++;
      if (sscanf(pbuf, "%d", &MU_up_tmp) != 1) {
	Eprint(__FILE__, __LINE__);
	// Ksnet::Sub_connect
	// Cannot scan 'object number of "rpu"'.
	Oprint();
	exit(1);
      }
//************************************************************
      MU_up[segment] = new int[2];
//************************************************************
      rpu_NO[MU_up_tmp] = segment;
    }                               

    while(*pbuf != ':') pbuf++;
    pbuf++;

    if (sscanf(pbuf, "%d", &sigma_down_tmp) != 1) {
      Eprint(__FILE__, __LINE__);
      // Ksnet::Sub_connect
      // Cannot scan 'sigma_down'.
      Oprint();
      exit(1);
    }
    sigma_down[segment] = sigma_down_tmp;

    while(*pbuf != ':') pbuf++;
    pbuf++;
    
    if (sigma_down[segment] > 0){
      MU_down[segment] = new int[sigma_down[segment]];
      for(j = 0; j < sigma_down[segment]; j++){
	while((*pbuf == ' ') || (*pbuf == '\t')) pbuf++;
	if (sscanf(pbuf, "%d", &MU_down_tmp) != 1) {
	  Eprint(__FILE__, __LINE__);
	  // Ksnet::Sub_connect
	  // Cannot scan 'MU_down'.
	  Oprint();
	  exit(1);
	}
	MU_down[segment][j] = MU_down_tmp;  // MU_down ˤϼºݤβƻʬ
	                                    // ֹ椬Ǽ롣

	while ((*pbuf != ' ') && (*pbuf != '\t')) pbuf++;
      }
    }

    else {
      while((*pbuf == ' ') || (*pbuf == '\t')) pbuf++;
      if (sscanf(pbuf, "%d", &MU_down_tmp) != 1) {
	Eprint(__FILE__, __LINE__);
	// Ksnet::Sub_connect
	// Cannot scan 'object number of "sp"'.
	Oprint();
	exit(1);
      }
      sp_NO[MU_down_tmp] = segment;
//************************************************************
      MU_down[segment] = new int[2];
//************************************************************
    }                                     
  }
  fclose(fp_c);
  delete [] buf;

  if (segment != nsegment - 1) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Sub_connect
    // Too few information for sub-connect.
    Oprint();
    exit(1);
  }

  /* mu_up, mu_down  */
  int k;
  for(i = 0; i < nsegment; i++){
    if(sigma_up[i] > 0){
      mu_up[i] = new int[sigma_up[i]];
      for(j = 0; j < sigma_up[i]; j++){
	for(k = 0; k < nsegment; k++){
	  if(abs(MU_up[i][j]) == seg_NO[k]){
	    if(MU_up[i][j] > 0) mu_up[i][j] = k+1;  // |k+1|򲾤βƻʬֹ
	    else mu_up[i][j] = -(k+1);              // Ȥ롣
	    break;
	  }
	}
      }
    }
    else if (sigma_up[i] == 0) {
      mu_up[i] = new int[2];
    }
    if(sigma_down[i] > 0){
      mu_down[i] = new int[sigma_down[i]];
      for(j = 0; j < sigma_down[i]; j++){
	for(k = 0; k < nsegment; k++){
	  if(abs(MU_down[i][j]) == seg_NO[k]){
	    if(MU_down[i][j] > 0) mu_down[i][j] = k+1; // |k+1|򲾤βƻʬ
	    else mu_down[i][j] = -(k+1);               // ֹȤ롣
	    break;
	  }
	}
      }
    }
    else if (sigma_down[i] == 0) {
      mu_down[i] = new int[2];
    }
  }

  for(i = 0; i < nsegment; i++){
    delete [] MU_up[i];
    delete [] MU_down[i];
  }
  delete [] MU_up;
  delete [] MU_down;

  /* ׻η */

  Boolean* cal_order_flag;  // βƻʬη׻˷ꤵƤ
                            // ɤ򼨤եåѿǼ.
  Boolean FLAG;        // ׻η꤬ǽɤ򼨤եåѿ.

  int n = 0;

  cal_order_flag = new Boolean[nsegment];

  for(i = 0; i < nsegment; i++){     // Ƥ cal_order_flag  NO 
    cal_order_flag[i] = NO;          // .
  }
  
  for(i = 0; i < nsegment; i++){
    if(sigma_up[i] == 0){                
      cal_order[n] = i;         // n+1 ܤ˷׻ƻʬֹ 
                                  // i+1 Ǥ롣
      cal_order_flag[i] = YES;  // ֹդνäƻʬΥեåѿ
                                // YES ˤ롣

//      printf("seg_NO = %d, cal_order[%d] = %d \n",
//	     seg_NO[cal_order[n]], n, cal_order[n]);
      n++;
    }
  }
  
  while (n < nsegment){
    for(i = 0; i < nsegment; i++){
      if(cal_order_flag[i] == YES) continue;
      FLAG = YES;
      for(j = 0; j < sigma_up[i]; j++){
	if((mu_up[i][j] > 0) 
	   && (cal_order_flag[mu_up[i][j] - 1] == NO)){
	  FLAG = NO;
	  break;
	}
      }
      
      if(FLAG == YES){ 
	cal_order[n] = i;
	cal_order_flag[i] = YES;

//	printf("seg_NO = %d, cal_order[%d] = %d \n",
//	       seg_NO[cal_order[n]], n, cal_order[n]);

	n++;
      }
    }
  }
  delete [] cal_order_flag;
}


double Ksnet::Storage_by_continuity(double q0_tmp)
{
  double storage;
  storage = storage_current_time + 
    (sum_input_current + lateral_current_time
     + lateral_next_time - sum_output_current - sum_length * q0_tmp)
      * time_step / 2.;
  return storage;
}


double Ksnet::Storage_by_integral(double q0_tmp)
{
  double storage = 0.;
  double s_tmp;
  double sum_local_input;                //  ʬήؤή̤
                                         //              ^^^^^^
  int i, j, k;

  for (i = 0; i < nsegment; i++){
    sum_local_input = 0;
    j = cal_order[i];

    if (sigma_up[j] != 0){
      for (k = 0; k < sigma_up[j]; k++){
	if (mu_up[j][k] > 0){
	  sum_local_input += Q_down[mu_up[j][k] - 1];
	}
      }
      Q_up[j] = alpha[j] * sum_local_input;
    }
    Q_down[j] = Q_up[j] + q0_tmp * L[j];

    if (q0_tmp == 0.) {
      s_tmp = K[j] * L[j] * pow(Q_up[j], P[j]);
    }
    else {
      s_tmp = K[j] / q0_tmp / (P[j] + 1.)
	* (pow((fabs(Q_down[j])), (P[j] + 1.))
	   - pow((fabs(Q_up[j])), (P[j] + 1.)));
    }
    storage += s_tmp;
  }
  return storage;
}


Boolean Ksnet::Initial_output(void)
{
  /* q0 η׻ */
  if(sum_length == 0.0) q0 = 0.0;
  else q0 = (sum_output_current - sum_input_current) / sum_length;

  /* α̡ή̤η׻ */
  int i;

  storage_current_time = Storage_by_integral(q0);
  sum_output_current = 0;
  for (i = 0; i < noutput; i++){
    sum_output_current += Q_down[sp_NO[i]];
    
    /* (m^3/secǽ) */
    sp[i].Send_data(current_time, Q_down[sp_NO[i]]);
  }
  return YES;
}


Boolean Ksnet::Calculate(void)
{
  int n;                                        /* ʬˡȿ */
  
  double q0_upper;                                     /* q0 ξ */
  double q0_lower;                                     /* q0 β */
  double q0_half;                  /* q0_upper  q0_lower Ȥʿ */
  double S_half;                          /* q0_half Ǥα̤κ */
  double S_lower;                        /* q0_lower Ǥα̤κ */
  
  /* ήüή */
  sum_input_next = 0.;
  int i;
  for (i = 0; i < ninput; i++) {
    Q_up[rpu_NO[i]] = rpu[i].Get_data(current_time + time_step);
    sum_input_next += Q_up[rpu_NO[i]];
  }
  
  /* ¦ή */
  lateral_next_time = 0.;
  for (i = 0; i < nlateral; i++) {
    lateral_next_time +=
      rpl[i].Get_data(current_time + time_step);

    /* ¦ήñ̤ m^2/s ΤȤ */
    if(lateral_unit == 2){
      lateral_next_time *= sum_length;
    }
  }

  /* q0 ξ¡¤ */
  if(sum_length == 0.0){
    q0_upper = 0.0;
    q0_lower = 0.0;
  }

  else{
    q0_upper = (2. * storage_current_time / time_step 
		+ sum_input_current + lateral_current_time + 
		lateral_next_time - sum_output_current) / sum_length;
    
    q0_lower = ((storage_current_time - Storage_by_integral(q0_upper))
		* 2. / time_step + sum_input_current + lateral_current_time
		+ lateral_next_time - sum_output_current) / sum_length;
  }

  /* ʬˡˤ q0 η׻ */
  n = 0;
  do {
    n++;
    q0_half = (q0_upper + q0_lower) / 2.;
    
    S_half = Storage_by_integral(q0_half)
      - Storage_by_continuity(q0_half);
    S_lower = Storage_by_integral(q0_lower)
      - Storage_by_continuity(q0_lower);
    
    if ((S_half * S_lower) < 0.) {
      q0_upper = q0_half;
    }
    else if ((S_half * S_lower) > 0.) {
      q0_lower = q0_half;
    }
    else if (S_half == 0.) {
      q0_upper = q0_half;
      q0_lower = q0_half;
    }
    else if (S_lower == 0.) {
      q0_upper = q0_lower;
    }
  }
  while((n < 300) && (fabs(q0_upper - q0_lower) > 0.000001));
  if (n >= 300) {
    Eprint(__FILE__, __LINE__);
    // Ksnet::Calculate
    // Failure to converge.
    Oprint();
    return NO;
  }
  
  q0 = (q0_upper + q0_lower) / 2.;

  storage_current_time = Storage_by_integral(q0);

  sum_output_next = 0;
  for(i = 0; i < noutput; i++){
    if (Q_down[sp_NO[i]] < 0.) {
      Q_down[sp_NO[i]] = 0.;
    }
    sum_output_next += Q_down[sp_NO[i]]; 

    /* m^3/s ǽ */
    sp[i].Send_data(current_time + time_step, Q_down[sp_NO[i]]);

//    printf("object_name : %s, current_time = %d,  Q_down = %f\n",
//	    object_name, current_time, Q_down[sp_NO[i]]);

  }
  
  /* ̤ι */
  sum_output_current = sum_output_next;
  sum_input_current = sum_input_next;
  lateral_current_time = lateral_next_time;
   
//  printf("'%s %d' %ld : %lf -> %lf\n", object_name, object_number,
//	 Next_time(), sum_input_next, sum_output_next);

//  printf("object_name : %s, current_time = %d \n",
//	 object_name, current_time);
  
  return YES;
}
